/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
 *  Copyright (C) 1999	TAKAI Kousuke  <tak@kmc.kyoto-u.ac.jp>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "shared.h"

#include "filesys.h"

typedef unsigned int	__u8	__attribute__((mode(QI)));
typedef unsigned int	__u16	__attribute__((mode(HI)));
typedef unsigned int	__u32	__attribute__((mode(SI)));

#include "fat.h"

static int num_clust;
static int mapblock;
static int data_offset;
static int fat_size;
static int sectors_per_record;
static int root_cluster;
static int root_offset;
static int root_length;

/* pointer(s) into filesystem info buffer for DOS stuff */
#define BPB	((struct fat_boot_sector *) (FSYS_BUF + 0x7E00))
#define FAT_BUF ( FSYS_BUF + 30208 )  /* 4 sector FAT buffer */
#define	FAT_BUF_SIZE	2048

int
fat_mount(void)
{
  if ((!ENTIRE_DISK_P(current_partition)
       && !PC_SLICE_TYPE_CHECK(current_slice, FAT12)
       && !PC_SLICE_TYPE_CHECK(current_slice, FAT16_LT32M)
       && !PC_SLICE_TYPE_CHECK(current_slice, FAT16_GT32M)
       && !PC_SLICE_TYPE_CHECK(current_slice, FAT32)
       && !BSD_FS_TYPE_CHECK(current_slice, MSDOS))
      || !devread(0, 0, 128, BPB)
      || (BPB->bytes_per_sector & (SECTOR_SIZE(buf_geom) - 1)) != 0
      || ((sectors_per_record = TO_NR_SECTORS(buf_geom, BPB->bytes_per_sector))
	  > 16)
      || (sectors_per_record & (sectors_per_record - 1)) != 0
      || BPB->sectors_per_cluster < 1 || BPB->sectors_per_cluster > 64
      || (BPB->sectors_per_cluster & (BPB->sectors_per_cluster - 1)) != 0)
    return 0;
  
  if (BPB->sectors_per_fat)
    {
      if (((__u8 *) BPB)[38] != FAT_BPB_SIGNATURE)
	return 0;

      /* FAT 12 & 16 */
      root_offset = (BPB->reserved_sectors
		     + BPB->sectors_per_fat * BPB->number_of_fats);
      root_length = TO_NR_SECTORS(buf_geom,
				  BPB->root_entries
				  * sizeof(struct fat_dir_entry)
				  + SECTOR_SIZE(buf_geom) - 1);
      data_offset = root_offset + root_length;
      root_cluster = 0;
      fat_size = 3;
    }
  else
    {
      if (((__u8 *) BPB)[66] != FAT_BPB_SIGNATURE)
	return 0;

      /* FAT 32 */
      data_offset = (BPB->reserved_sectors
		     + BPB->big_sectors_per_fat * BPB->number_of_fats);
      root_cluster = BPB->root_dir_start_cluster;
      fat_size = 8;
    }

  num_clust = (((BPB->total_sectors
		 ? BPB->total_sectors : BPB->big_total_sectors)
		- data_offset) / BPB->sectors_per_cluster) + 2;
  if (fat_size == 3 && num_clust > FAT_MAX_12BIT_CLUST)
    fat_size++;

  /* extra check for floppy */
  if (!DRIVE_HARDDISK_P(current_drive)
      && (BPB->big_total_sectors != 0
	  || num_clust < FAT_MIN_NUM_SECTORS
	  || (NSECTORS_SIZE(buf_geom, BPB->sectors_per_fat) * 2 / 3
	      <= num_clust)))
    return 0;

  DPRINTF(DEBUG_FILESYSTEM, "[fat: %d clusters, %d bits FAT]",
	  num_clust, fat_size * 4);
  mapblock = -4096;

  return 1;
}


static int
fat_create_blocklist(int first_fat_entry)
{
  BLK_CUR_FILEPOS = 0;
  BLK_CUR_BLKNUM = 0;
  BLK_CUR_BLKLIST = BLK_BLKLIST_START;
  block_file = 1;
  filepos = 0;

  if (first_fat_entry == 0)
    {
      /* root directory */

      BLK_BLKSTART(BLK_BLKLIST_START) = root_offset;
      BLK_BLKLENGTH(BLK_BLKLIST_START) = root_length;
      fsmax = filemax = sizeof(struct fat_dir_entry) * BPB->root_entries;
    }
  else    /* any real directory/file */
    {
      int blk_cur_blklist = BLK_BLKLIST_START;
      int last_fat_entry, new_mapblock;

      fsmax = 0;

      do
	{
	  BLK_BLKSTART(blk_cur_blklist)
	    = (((first_fat_entry - 2) * BPB->sectors_per_cluster + data_offset)
	       * sectors_per_record);
	  BLK_BLKLENGTH(blk_cur_blklist) = 0;

	  do
	    {
	      BLK_BLKLENGTH(blk_cur_blklist) += FAT_BPB_SECT_PER_CLUST(BPB);
	      last_fat_entry = first_fat_entry;

	      /*
	       *  Do FAT table translation here!
	       */

	      new_mapblock = (last_fat_entry * fat_size) >> 1;
	      if (new_mapblock > (mapblock + 2045)
		  || new_mapblock < (mapblock + 3))
		{
		  mapblock = ((new_mapblock < 6) ? 0
			      : ((new_mapblock - 6) & -SECTOR_SIZE(buf_geom)));
		  if (!devread(TO_NR_SECTORS(buf_geom, mapblock)
			       + BPB->reserved_sectors * sectors_per_record,
			       0, FAT_BUF_SIZE, FAT_BUF))
		    return 0;
		}

	      first_fat_entry
		= *((__u32 *) (FAT_BUF + (new_mapblock - mapblock)));

	      if (fat_size == 3)
		{
		  if (last_fat_entry & 1)
		    first_fat_entry >>= 4;
		  first_fat_entry &= 0xFFF;
		}
	      else if (fat_size == 4)
		first_fat_entry &= 0xFFFF;

	      if (first_fat_entry < 2)
		{
		  errnum = ERR_FSYS_CORRUPT;
		  return 0;
		}
	    }
	  while (first_fat_entry == (last_fat_entry + 1)
		 && first_fat_entry < num_clust);

	  BLK_BLKLENGTH(blk_cur_blklist) *= sectors_per_record;
	  fsmax += NSECTORS_SIZE(buf_geom, BLK_BLKLENGTH(blk_cur_blklist));
	  blk_cur_blklist += BLK_BLKLIST_INC_VAL;
	}
      while (first_fat_entry < num_clust && blk_cur_blklist < (FAT_BUF - 7));
    }

  DPRINTF(DEBUG_FILESYSTEM, "[fat: fsmax %u]", fsmax);
  return 1;
}


/* XX FAT filesystem uses the block-list filesystem read function,
   so none is defined here.  */


int
fat_dir(char *dirname)
{
  char *rest, ch, filename[13], dir_buf[FAT_DIRENTRY_LENGTH];
  int attrib = FAT_ATTRIB_DIR, map = root_cluster;

/* main loop to find desired directory entry */
loop:

  if (!fat_create_blocklist(map))
    return 0;

  /* if we have a real file (and we're not just printing possibilities),
     then this is where we want to exit */

  if (!*dirname || isspace(*dirname))
    {
      if (attrib & FAT_ATTRIB_DIR)
	{
	  errnum = ERR_BAD_FILETYPE;
	  return 0;
	}

      return 1;
    }

  /* continue with the file/directory name interpretation */

  while (*dirname == '/')
    dirname++;

  filemax = fsmax;

  if (!filemax || !(attrib & FAT_ATTRIB_DIR))
    {
      errnum = ERR_BAD_FILETYPE;
      return 0;
    }

  for (rest = dirname; (ch = *rest) && !isspace(ch) && ch != '/'; rest++) ;

  *rest = 0;

  do
    {
      if (read(dir_buf, FAT_DIRENTRY_LENGTH) != FAT_DIRENTRY_LENGTH)
	{
	  if (!errnum)
	    {
	      if (print_possibilities < 0)
		{
		  putchar('\n');
		  return 1;
		}

	      errnum = ERR_FILE_NOT_FOUND;
	      *rest = ch;
	    }

	  return 0;
	}

      if (!FAT_DIRENTRY_VALID(dir_buf))
	continue;

      /* XXX convert to 8.3 filename format here */
      {
	int i, j, c;

	for (i = 0; i < 8 && (c = filename[i] = tolower(dir_buf[i]))
	       && !isspace(c) ; i++) ;

	filename[i++] = '.';

	for (j = 0; j < 3 && (c = filename[i+j] = tolower(dir_buf[8+j]))
	       && !isspace(c) ; j++) ;

	if (j == 0)
	  i--;

	filename[i+j] = 0;
      }

      if (print_possibilities && ch != '/'
	  && (!*dirname || strcmp(dirname, filename) <= 0))
	{
	  if (print_possibilities > 0)
	    print_possibilities = -print_possibilities;
#ifdef PC9800
	  putchar_flags = PUTCHAR_SJIS;
#endif
	  printf("  %s", filename);
#ifdef PC9800
	  putchar_flags = 0;
#endif
	}
    }
  while (strcmp(dirname, filename) != 0 || (print_possibilities && ch != '/'));

  *(dirname = rest) = ch;

  attrib = FAT_DIRENTRY_ATTRIB(dir_buf);
  filemax = FAT_DIRENTRY_FILELENGTH(dir_buf);
  map = FAT_DIRENTRY_FIRST_CLUSTER(dir_buf);

  /* go back to main loop at top of function */
  goto loop;
}
