/*
 *  ISO 9660 filesystem backend for GRUB (GRand Unified Bootloader)
 *  including Rock Ridge Extensions support
 *
 *  Copyright (C) 1998, 1999  Kousuke Takai  <tak@kmc.kyoto-u.ac.jp>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 *  References:
 *	linux/fs/isofs/rock.[ch]
 *	mkisofs-1.11.1/diag/isoinfo.c
 *	mkisofs-1.11.1/iso9660.h
 *		(all are written by Eric Youngdale)
 */

#ifdef	TEST
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>

extern int iso9660_mount(void);
extern int iso9660_dir(char *);
extern int iso9660_read(int, int);

char	print_possibilities = 0;
int	errnum;
int	filemax;
int	fsmax;
int	filepos;
char	block_file;

#define	ERR_FSYS_CORRUPT	1
#define	ERR_FILE_NOT_FOUND	2
#define	ERR_BAD_FILETYPE	3
#define	ERR_READ		4

int devread(int, int, int, int);

#ifdef PC9800
struct disk_geometry {
	unsigned char	log2_sector_size;
	unsigned char	da_ua;
	unsigned char	sectors;
	unsigned char	heads;
	unsigned int	total_sectors;
};
# define SECTOR_SIZE_SHIFT(geom)	((geom)->log2_sector_size)

static struct disk_geometry disk_geometry = { 9, 0x00, 0, 0, ~0 };
struct disk_geometry *buf_geom = &disk_geometry;
#else
# define SECTOR_SIZE_SHIFT(geom)	9
#endif

long current_partition = 0xFFFFFF;
long current_slice = 0;
unsigned char debug_flags = ~0;

unsigned char FSYS_BUF[32768];

int BLK_CUR_FILEPOS;
int BLK_CUR_BLKNUM;
int BLK_CUR_BLKLIST;
struct blk_blklist {
	int start, length;
};
struct blk_blklist blk_blklist[2];
#define BLK_BLKLIST_START	((int)blk_blklist)
#define BLK_BLKSTART(list)	((struct blk_blklist *)(list)->start)
#define BLK_BLKLENGTH(list)	((struct blk_blklist *)(list)->length)
#define BLK_BLKLIST_INC_VAL	(sizeof(struct blk_blklist))

#define	DEBUG
#define	DPRINTF(flag, fmt, arg...)	fprintf(stderr, fmt , ## arg)

#else	/* !TEST */

#include "shared.h"
#include "filesys.h"

#endif	/* !TEST */

#ifndef	__BIT_TYPES_DEFINED__
typedef		 int	 int8_t	__attribute__((mode(QI)));
typedef unsigned int   u_int8_t	__attribute__((mode(QI)));
typedef		 int	int16_t	__attribute__((mode(HI)));
typedef unsigned int  u_int16_t	__attribute__((mode(HI)));
typedef		 int	int32_t	__attribute__((mode(SI)));
typedef unsigned int  u_int32_t	__attribute__((mode(SI)));
#endif

#define	PRIMDESC	((struct iso_primary_descriptor *)(FSYS_BUF + 2048))
#define	DIRREC		((struct iso_directory_record *)(FSYS_BUF + 4096))
#define	RRCONTBUF	((unsigned char *)(FSYS_BUF + 6144))
#define	RRNAMEBUF	((unsigned char *)(FSYS_BUF + 8192))

#define	ISO_REGULAR	1	/* regular file	*/
#define	ISO_DIRECTORY	2	/* directory	*/
#define	ISO_OTHER	0	/* other file (with Rock Ridge) */

#define	RR_FLAG_PX	0x01	/* have POSIX file attributes */
#define	RR_FLAG_NM	0x08	/* have alternate file name   */

/* POSIX file attributes for Rock Ridge extensions */
#define	POSIX_S_IFMT	0xF000
#define	POSIX_S_IFREG	0x8000
#define	POSIX_S_IFDIR	0x4000

/* volume descriptor types */
#define ISO_VD_PRIMARY 1
#define ISO_VD_END 255

#define ISO_STANDARD_ID "CD001"

#define	__PACKED_STRUCT	__attribute__((__packed__))

struct __iso_16bit {
	u_int16_t l, b;
} __PACKED_STRUCT;

struct __iso_32bit {
	u_int32_t l, b;
} __PACKED_STRUCT;

typedef	union { u_int8_t l,b; }	iso_8bit_t;
typedef	struct __iso_16bit	iso_16bit_t;
typedef	struct __iso_32bit	iso_32bit_t;
typedef u_int8_t		iso_date_t[7];

struct iso_directory_record {
	iso_8bit_t	length;
	iso_8bit_t	ext_attr_length;
	iso_32bit_t	extent;
	iso_32bit_t	size;
	iso_date_t	date;
	iso_8bit_t	flags;
	iso_8bit_t	file_unit_size;
	iso_8bit_t	interleave;
	iso_16bit_t	volume_seq_number;
	iso_8bit_t	name_len;
	u_int8_t	name[1];
} __PACKED_STRUCT;

struct iso_primary_descriptor {
	iso_8bit_t	type;
	u_int8_t	id[5];
	iso_8bit_t	version;
	u_int8_t	_unused1[1];
	u_int8_t	system_id[32];
	u_int8_t	volume_id[32];
	u_int8_t	_unused2[8];
	iso_32bit_t	volume_space_size;
	u_int8_t	_unused3[32];
	iso_16bit_t	volume_set_size;
	iso_16bit_t	volume_seq_number;
	iso_16bit_t	logical_block_size;
	iso_32bit_t	path_table_size;
	u_int8_t	type_l_path_table[4];
	u_int8_t	opt_type_l_path_table[4];
	u_int8_t	type_m_path_table[4];
	u_int8_t	opt_type_m_path_table[4];
	struct iso_directory_record root_directory_record;
	u_int8_t	volume_set_id[128];
	u_int8_t	publisher_id[128];
	u_int8_t	preparer_id[128];
	u_int8_t	application_id[128];
	u_int8_t	copyright_file_id[37];
	u_int8_t	abstract_file_id[37];
	u_int8_t	bibliographic_file_id[37];
	u_int8_t	creation_date[17];
	u_int8_t	modification_date[17];
	u_int8_t	expiration_date[17];
	u_int8_t	effective_date[17];
	iso_8bit_t	file_structure_version;
	u_int8_t	_unused4[1];
	u_int8_t	application_data[512];
	u_int8_t	_unused5[653];
} __PACKED_STRUCT;

struct rock_ridge {
	u_int16_t	signature;
	u_int8_t	len;
	u_int8_t	version;
	union {
	  struct CE {
	    iso_32bit_t	extent;
	    iso_32bit_t	offset;
	    iso_32bit_t	size;
	  } ce;
	  struct NM {
	    iso_8bit_t	flags;
	    u_int8_t	name[0];
	  } nm;
	  struct PX {
	    iso_32bit_t	mode;
	    iso_32bit_t	nlink;
	    iso_32bit_t	uid;
	    iso_32bit_t	gid;
	  } px;
	  struct RR {
	    iso_8bit_t	flags;
	  } rr;
	} u;
} __PACKED_STRUCT;

union RR_ptr {
	struct rock_ridge * rr;
	void *		    ptr;
	int		    i;
};

typedef	union RR_ptr	RR_ptr_t;

#define	RRMAGIC(c1, c2)	((c1)|(c2) << 8)

#define	offsetof(TYPE, MEMBER) ((char *)&(((TYPE *)0)->MEMBER) - (char *)0)

#ifndef TEST_ASM

static inline
int iso9660_devread(unsigned sector, unsigned size, void * addr)
{
#if 0    /* GCC 2.8 will generate redundant code to compute shift count. */
  sector <<= (u_int8_t)((u_int8_t)11 - log2_sector_size);
#else
  __asm__("shl%L0 %1,%0"
	  : "=r"(sector)
	  : "Ic"((int8_t)((int8_t)11 - SECTOR_SIZE_SHIFT(buf_geom))),
	    "0"(sector));
#endif
  return devread(sector, 0, size, (int)addr);
}

#define	CHECK2(ptr, c1, c2) \
	(*(unsigned short *)(ptr) == (((c1) | (c2) << 8) & 0xFFFF))
#define	CHECK4(ptr, c1, c2, c3, c4) \
	(*(unsigned long *)(ptr) == ((c1) | (c2)<<8 | (c3)<<16 | (c4)<<24))

int
iso9660_mount(void)
{
  unsigned int sector;

  /*
   *  Because there is no defined slice type ID for ISO-9660 filesystem,
   *  this test will pass only either (1) if entire disk is used, or
   *  (2) if current partition is BSD style sub-partition whose ID is
   *  ISO-9660.
   */
  if (!ENTIRE_DISK_P(current_partition)
      && !BSD_FS_TYPE_CHECK(current_slice, ISO9660))
    return 0;

  for (sector = 16 ; sector < 32 ; sector++)
    {
      if (!iso9660_devread(sector, sizeof(*PRIMDESC), PRIMDESC)) 
	return 0;
      /* check ISO_VD_PRIMARY and ISO_STANDARD_ID */
      if (CHECK4(&PRIMDESC->type, ISO_VD_PRIMARY, 'C', 'D', '0')
	  && CHECK2(PRIMDESC->id + 3, '0', '1'))
	{
	  DPRINTF(DEBUG_FILESYSTEM,
		  "iso9660: primary descriptor found at sector %u\n", sector);
	  return 1;
	}
    }
  return 0;
}

int iso9660_dir(const char * dirname)
{
  struct iso_directory_record *idr;

  idr = &PRIMDESC->root_directory_record;

  do
    {
      unsigned int pathlen;
      int size;
      unsigned int extent;

      while (*dirname == '/')	/* skip leading slashes */
	dirname++;
      /* pathlen = strcspn(dirname, "/\n\t "); */
      for (pathlen = 0 ;
	   dirname[pathlen]
	     && !isspace(dirname[pathlen]) && dirname[pathlen] != '/' ;
	   pathlen++)
	;

      size = idr->size.l;
      extent = idr->extent.l;

      while (size > 0)
	{
	  if (!iso9660_devread(extent, 2048, DIRREC))
	    {
	      errnum = ERR_FSYS_CORRUPT;
	      return 0;
	    }
	  extent++;

	  idr = (struct iso_directory_record *)DIRREC;
	  while (idr->length.l > 0)
	    {
	      const char * name = idr->name;
	      unsigned int name_len = idr->name_len.l;
	      RR_ptr_t rr_ptr;
	      struct rock_ridge * ce_ptr;
	      unsigned int rr_len;
	      unsigned char file_type;
	      unsigned char rr_flag;

	      file_type = (idr->flags.l & 2) ? ISO_DIRECTORY : ISO_REGULAR;
	      if (name_len == 1)
		{
		  static const char dotdot[2] = "..";

		  if (name[0] == 0)
		    name = dotdot + 1;
		  else if (name[0] == 1)
		    {
		      name = dotdot;
		      name_len = 2;
		    }
		}
	      if (name_len > 2 && CHECK2(name + name_len - 2, ';', '1'))
		{
		  name_len -= 2;	/* truncate trailing file version */
		  if (name_len > 1 && name[name_len - 1] == '.')
		    name_len--;		/* truncate trailing dot */
		}

	      /*
	       *  Parse Rock-Ridge extension
	       */
	      rr_len = (idr->length.l - idr->name_len.l
			- (unsigned char)sizeof(struct iso_directory_record)
			+ (unsigned char)sizeof(idr->name));
	      rr_ptr.ptr = ((unsigned char *)idr + idr->name_len.l
			    + sizeof(struct iso_directory_record)
			    - sizeof(idr->name));
	      if (rr_ptr.i & 1)
		rr_ptr.i++, rr_len--;
	      ce_ptr = NULL;
	      rr_flag = RR_FLAG_NM | RR_FLAG_PX;

	      while (rr_len >= 4)
		{
		  if (rr_ptr.rr->version != 1)
		    DPRINTF(DEBUG_FILESYSTEM,
			    "Non-supported version (%d) RockRidge chunk "
			    "`%c%c'\n", rr_ptr.rr->version,
			    rr_ptr.rr->signature & 0xFF,
			    rr_ptr.rr->signature >> 8);
		  else if (rr_ptr.rr->signature == RRMAGIC('R', 'R')
			   && rr_ptr.rr->len >= 5)
		    rr_flag &= rr_ptr.rr->u.rr.flags.l;
		  else if (rr_ptr.rr->signature == RRMAGIC('N', 'M'))
		    {
		      name = rr_ptr.rr->u.nm.name;
		      name_len = rr_ptr.rr->len - 5;
		      rr_flag &= ~RR_FLAG_NM;
		    }
		  else if (rr_ptr.rr->signature == RRMAGIC('P', 'X')
			   && rr_ptr.rr->len >= 36)
		    {
		      file_type = ((rr_ptr.rr->u.px.mode.l & POSIX_S_IFMT)
				   == POSIX_S_IFREG
				   ? ISO_REGULAR
				   : ((rr_ptr.rr->u.px.mode.l & POSIX_S_IFMT)
				      == POSIX_S_IFDIR
				      ? ISO_DIRECTORY : ISO_OTHER));
		      rr_flag &= ~RR_FLAG_PX;
		    }
		  else if (rr_ptr.rr->signature == RRMAGIC('C', 'E')
			   && rr_ptr.rr->len >= 28)
		    ce_ptr = rr_ptr.rr;
		  if (!rr_flag)
		    /*
		     * There is no more extension we expects...
		     */
		    break;
		  rr_len -= rr_ptr.rr->len;
		  rr_ptr.ptr += rr_ptr.rr->len;
		  if (rr_len < 4 && ce_ptr != NULL)
		    {
		      /* preserve name before loading new extent. */
		      if( (void *)RRCONTBUF <= (void *)name
			  && (void *)name < (void *)RRCONTBUF + 2048 )
			{
			  /* `__builtin_memcpy' calls `memcpy'. */
			  __asm__("cld\n\trep\n\tmovsb"
				  : /* no output */
				  : "S"(name), "D"(RRNAMEBUF), "c"(name_len)
				  : "esi", "edi", "ecx", "memory");
			  name = RRNAMEBUF;
			}
		      rr_ptr = RRCONTBUF + ce_ptr->u.ce.offset.l;
		      rr_len = ce_ptr->u.ce.size.l;
		      if (!iso9660_devread(ce_ptr->u.ce.extent.l, 2048,
					   RRCONTBUF))
			{
			  errnum = 0;	/* this is not fatal. */
			  break;
			}
		      ce_ptr = NULL;
		    }
		}

	      if (name_len >= pathlen
		  && !__builtin_memcmp(name, dirname, pathlen))
		{
		  if (dirname[pathlen] == '/' || !print_possibilities)
		    {
		      /*
		       *  DIRNAME is directory component of pathname,
		       *  or we are to open a file.
		       */
		      if (pathlen == name_len)
			{
			  if (dirname[pathlen] == '/')
			    {
			      if (file_type != ISO_DIRECTORY)
				{
				  errnum = ERR_BAD_FILETYPE;
				  return 0;
				}
			      goto next_directory;
			    }
			  if (file_type != ISO_REGULAR)
			    {
			      errnum = ERR_BAD_FILETYPE;
			      return 0;
			    }
			  __asm__("shl%L0 %1,%0"
				  : "r"(BLK_BLKSTART(BLK_BLKLIST_START))
				  : "Ic"((int8_t)
					 ((int8_t)11
					  - SECTOR_SIZE_SHIFT(buf_geom))),
				    "0"(idr->extent.l));
			  fsmax = filemax = idr->size.l;
			  BLK_BLKLENGTH(BLK_BLKLIST_START)
			    = TO_NR_SECTORS(buf_geom, idr->size.l) + 1;
			  BLK_CUR_BLKLIST = BLK_BLKLIST_START;
			  BLK_CUR_BLKNUM = 0;
			  BLK_CUR_FILEPOS = 0;
			  return 1;
			}
		    }
		  else
		    {
		      print_possibilities = -1;
		      putchar(' ');
		      putchar(' ');
		      while (name_len)
			{
			  putchar(*name++);
			  name_len--;
			}
		    }
		}
	      idr = (void *)((char *)idr + idr->length.l);
	    }
	  size -= 2048;
	}
      if (dirname[pathlen] == '/' || print_possibilities >= 0)
	{
	  errnum = ERR_FILE_NOT_FOUND;
	  return 0;
	}
    next_directory:
      dirname += pathlen;
    }
  while (*dirname == '/');

  putchar('\n');
  return 1;
}
#endif /* !TEST_ASM */

#ifdef	TEST

static int dev_fd;

int devread(int sector, int offset, int len, int addr)
{
  if( lseek(dev_fd, (sector << SECTOR_SIZE_SHIFT(buf_geom)) + offset,
	    SEEK_SET) < 0
      || read(dev_fd, (void *)addr, len) < 0 )
    {
      perror("devread");
      errnum = ERR_READ;
      return 0;
    }
  return 1;
}

void dprintf(const char *fmt, ...)
{
  va_list ap;

  fprintf(stderr, "debug: ");
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}

int main(int argc, char *argv[])
{
    int i;
    int do_read = 0;
    const char * dev = "/dev/sdb";

    while( (i = getopt(argc, argv, "d:pr")) >= 0 )
	switch( (char)i ) {
	case 'd':
	    dev = optarg;
	    break;
	case 'p':
	    print_possibilities = 1;
	    break;
	case 'r':
	    do_read = 1;
	    break;
	default:
	    return 1;
	}

    if( optind != argc - 1 ) {
	fprintf(stderr, "Usage: %s [-d DEVICE] [-p] [-r] PATH\n", argv[0]);
	return 1;
    }

    if( (i = open(dev, O_RDONLY)) < 0 ) {
	perror(dev);
	return 1;
    }
    dev_fd = i;

    if( !iso9660_mount() ) {
	fprintf(stderr, "%s: Cannot find ISO-9660 primary descriptor\n", dev);
	close(dev_fd);
	return 1;
    }

    i = iso9660_dir(argv[optind]);
    fflush(stdout);
    if( !i ) {
	switch( errnum ) {
	case ERR_BAD_FILETYPE:
	    fprintf(stderr, "%s: Bad file type\n", argv[optind]);
	    break;
	case ERR_FILE_NOT_FOUND:
	    fprintf(stderr, "%s: Not found\n", argv[optind]);
	    break;
	case ERR_FSYS_CORRUPT:
	    fprintf(stderr, "File system is corrupted\n");
	    break;
	case ERR_READ:
	    fprintf(stderr, "Read error\n");
	    break;
	default:
	    fprintf(stderr, "Unknown error %d\n", errnum);
	}
	return 1;
    }
    if( !print_possibilities ) {
	if( do_read ) {
#if 0
	    static char buf[1024];

	    for( filepos = 0 ; filepos < filemax ; ) {
		int size = (filemax - filepos) > 1024
		    ? 1024 : filemax - filepos;
		if( !iso9660_read((int)buf, size) ) {
		    fprintf(stderr, "iso9660_read() returned zero\n");
		    break;
		}
		write(1, buf, size);
	    }
#endif
	}
    }
    close(dev_fd);

    return 0;
}
#endif	/* TEST */

/*
 * Local variables:
 * compile-command: "gcc -O2 -Wall -DTEST -DPC9800 -fomit-frame-pointer -malign-functions=2 -malign-jumps=2 -malign-loops=2 -Wa,-a=/tmp/fsys_iso9660.list -o /tmp/fsys_iso9660.o -c fsys_iso9660.c"
 * compile-command: "gcc -O2 -Wall -DTEST -DTEST_ASM -DPC9800 fsys_iso9660.c /tmp/fsys_iso9660.o"
 * End:
 */
