/* ptable98.c -- GRUB/98 builtin partition table utility
   Copyright (C) 2001,2002  TAKAI Kousuke  <tak@kmc.gr.jp>

   This file is a part of the GRUB/98.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY and FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "shared.h"
#include "pc_slice.h"

#define PAGEBUF		((char *) FSYS_BUF)
#define PAGESIZE	4096

#define DEFINE_INTTYPE(Type, Unsigned, Mode) \
	typedef Unsigned int	Type __attribute__ ((__mode__ (Mode)))
DEFINE_INTTYPE (uint8,  unsigned, QI);
DEFINE_INTTYPE (uint16, unsigned, HI);
DEFINE_INTTYPE (uint32, unsigned, SI);
DEFINE_INTTYPE (uint64, unsigned, DI);
DEFINE_INTTYPE (int16, signed, HI);
DEFINE_INTTYPE (int32, signed, SI);

struct bios_disk_address
  {
    uint8 s, h;
    uint16 c;
  };

struct partition_descriptor
  {
    uint16 id;
    uint16 _unused;
    struct bios_disk_address ipl, start, end;
#define PARTITION_NAME_SIZE	16
    uint8 name[PARTITION_NAME_SIZE];
  };

struct pt_data
  {
    unsigned int n_partitions;
    unsigned int sectors_per_cylinder;
    unsigned int cylinders;
    char device_name[4];
    char edited;
    struct disk_geometry saved_geometry;
    struct partition_descriptor partition_table[0]
	__attribute__ ((aligned (sizeof (struct partition_descriptor))));
  };

#define DATA		(*(struct pt_data *) 0x100000)

/* Origin of partition index.  NEC's boot menu uses 1-based index,
   but we use 0 for consistency with other part of GRUB.  Sigh.  */
#define INDEX_BASE	0

#ifndef isblank
# define isblank(c)	((c) == ' ' || (c) == '\t')
#endif

/* Make up 32-bit word by concatenating bytes C0, C1, C2 and C3.  */
#define MAKE_U32(c0, c1, c2, c3)	((((c0) & 255U) <<  0) | \
					 (((c1) & 255U) <<  8) | \
					 (((c2) & 255U) << 16) | \
					 (((c3) & 255U) << 24))
#define MAKE_U16(c0, c1)	(((c0) & 255U) | (((c1) & 255U) << 8))

/* `(DIVIDEND + (DIVISOR - 1)) / DIVISOR' may overflow while using
   64-bit integral type for division requires GCC's runtime library.  */
#define CEIL_UDIV(dividend, divisor)				\
	({ unsigned int const _dividend = (dividend);		\
	   unsigned int const _divisor = (divisor);		\
	   unsigned int const _quadrant = _dividend / _divisor;	\
	   _dividend % _divisor ? _quadrant + 1 : _quadrant; })

static int
y_or_n_p (const char *prompt)
{
  uint8 saved_x;
  int yes;

  printf ("%s? (y/n) ", prompt);
  for (saved_x = putchar_x; ; putchar_x = saved_x)
    {
      int key = getkey ();
      if (' ' <= (char) key && (char) key <= '~')
	putchar ((char) key);
      else if ((char) key && !(key & 0x80))
	{
	  putchar ('^');
	  putchar ((char) key ^ 0x40); /* this can handle `^?' */
	}
      if ((char) key == 'Y' || (char) key == 'y' || key == KEY_ENTER)
	{
	  yes = 1;
	  break;
	}
      if ((char) key == 'N' || (char) key == 'n' || (char) key == 3 /* ^C */
	  || key == KEY_ESC)
	{
	  yes = 0;
	  break;
	}
    }
  putchar ('\n');
  return yes;
}  

static int
overwrite_ok (void)
{
  if (kernel_type == KERNEL_TYPE_PTEDITOR && DATA.edited /* && interactive */
      && !y_or_n_p (MSG2 ("Really throw away currently edited table", "\
ԽΥѡƥơ֥ޤǤ")))
    {
      printf (MSG2 ("Aborted.\n", "Ǥޤ\n"));
      return 0;
    }
  return 1;
}

static int
begun (void)
{
  if (kernel_type == KERNEL_TYPE_PTEDITOR)
    return 1;
  printf (MSG2 ("Target partition table not loaded\n", "\
оݤΥѡƥơ֥ɤƤޤ\n"));
  errnum = ERR_BAD_ARGUMENT;
  return 0;
}

#define ERROR_RETURN(e)	do { errnum = (e); return 1; } while (0)
#define BAD_ARGUMENT	ERROR_RETURN (ERR_BAD_ARGUMENT)

static int
pt_begin (char *arg)
{
  char *p;
  unsigned int n;
  const struct disk_geometry *geom;

  if (!(p = set_device (arg)))
    return 1;
  if (!DRIVE_HARDDISK_P (current_drive) || !ENTIRE_DISK_P (current_partition)
      || (*p && !isblank (*p)))
    ERROR_RETURN (ERR_DEV_VALUES);

  geom = get_geometry (current_drive);
  if (!GEOMETRY_DRIVE_EXIST_P (geom))
    ERROR_RETURN (ERR_NO_DISK);

  if (!GEOMETRY_CHS_AVAIL_P (geom))
    ERROR_RETURN (ERR_GEOM);

  if (!overwrite_ok ())
    return 1;

  /* Mark we are clobbering a loaded kernel if any.  */
  kernel_type = KERNEL_TYPE_NONE;

  if (!rawread (current_drive, 0, 0, 0, DATA.partition_table))
    return 1;
  if (*(uint16 *) ((char *) DATA.partition_table + SECTOR_SIZE (geom) - 2)
      != 0xAA55)
    printf (MSG2 ("Warning: partition table signature not found in MBR\n",
		  "ٹ: ѡƥơ֥뼱̻Ҥޤ\n"));
  if (!rawread (current_drive, 1, 0, 0, DATA.partition_table))
    return 1;

  DATA.saved_geometry = *geom;
  n = SECTOR_SIZE (geom) / sizeof (struct partition_descriptor);
  if (n > 16)	/* XXX ? */
    n = 16;
  DATA.n_partitions = n;
  DATA.sectors_per_cylinder = geom->heads * geom->sectors;
  if (DATA.sectors_per_cylinder == 0)
    ERROR_RETURN (ERR_GEOM);

  DATA.cylinders = TOTAL_SECTORS (geom) / DATA.sectors_per_cylinder;
  DATA.edited = 0;

  *(uint32 *) &DATA.device_name[0] = MAKE_U32 ('h', 'd', '0', '\0');
  if (GEOMETRY_SCSI_P (geom))
    DATA.device_name[0] = 's';
  DATA.device_name[2] = '0' + (geom->da_ua & 7);

  /* We are now using kernel loading area.  */
  kernel_type = KERNEL_TYPE_PTEDITOR;

  return 0;
}

static int
pt_end (char *arg)
{
  if (!overwrite_ok ())
    return 1;
  kernel_type = KERNEL_TYPE_NONE;
  return 0;
}

#define WRITE_BUF	FSYS_BUF

#define LINEAR_TO_SEGMENT(address)	((unsigned int) (address) >> 4)

static int
pt_write (char *arg)
{
  unsigned int size;
  int flags;
#define PW_IGNORE_GEOMETRY_CHANGE	1
#define PW_WRITE_SIGNATURE		2

  if (!begun ())
    return 1;
  flags = 0;
  size = SECTOR_SIZE (&DATA.saved_geometry);
 restart:
  if (!rawread (DATA.saved_geometry.da_ua, 0, 0, size * 2, WRITE_BUF))
    return 1;
  if (!(flags & PW_IGNORE_GEOMETRY_CHANGE)
      && __builtin_memcmp (get_geometry (DATA.saved_geometry.da_ua),
			   &DATA.saved_geometry,
			   sizeof (struct disk_geometry)))
    {
      if (!y_or_n_p (MSG2 ("\
Disk geometry changed during editing.  Continue", "\
Խ˥ǥȥ꤬ѲƤޤ񤭹ߤ³Ԥޤ")))
	{
	  printf (MSG2 ("Aborted.\n", "Ǥޤ\n"));
	  return 1;
	}
      flags |= PW_IGNORE_GEOMETRY_CHANGE;
      goto restart;
    }
  if (!(flags & PW_WRITE_SIGNATURE)
      && *(uint16 *) (WRITE_BUF + size - 2) != 0xAA55)
    {
      if (y_or_n_p (MSG2 ("No partition table signature on MBR.  Write it", "\
MBR ˥ѡƥơ֥뼱̻Ҥޤ󡣽񤭹ߤޤ")))
	flags |= PW_WRITE_SIGNATURE;
      goto restart;
    }
  *(uint16 *) (WRITE_BUF + size - 2) = 0xAA55;
  memcpy ((void *) (WRITE_BUF + size), DATA.partition_table, size);

  if (biosdisk (BIOSDISK_SUBFUNC_WRITE,
		get_geometry (DATA.saved_geometry.da_ua),
		(flags & PW_WRITE_SIGNATURE) ? 0 : 1,
		(flags & PW_WRITE_SIGNATURE) ? 2 : 1,
		LINEAR_TO_SEGMENT ((flags & PW_WRITE_SIGNATURE)
				   ? WRITE_BUF : WRITE_BUF + size)))
    ERROR_RETURN (ERR_WRITE);

  buf_sectors = 0;		/* Invalidate GRUB's track cache. */
  DATA.edited = 0;
  return 0;
}

static const uint8 known_type[] = ("\x20\x01" "FAT12\0"
				   "\x20\x11" "FAT16 (< 32M)\0"
				   "\x20\x21" "FAT16 (> 32M)\0"
				   "\x20\x61" "FAT32\0"
				   "\x20\xE2" "Linux native\0"
				   "\x20\x62" "Linux swap\0"
				   "\x94\x44" "BSD\0");

static inline const char *
partition_type_string (uint16 id)
{
  const uint8 *p = known_type;

  do
    if (*(const uint16 *) p == id)
      return p + 2;
    else
      for (++p; *++p; /* nop */);
  while (*++p);
  return NULL;
}

#define SPACEx4		MAKE_U32 (' ', ' ', ' ', ' ')

static int
empty_partition_p (const struct partition_descriptor *p)
{
  return (p->id == 0
	  && *(uint32 *) &p->name[ 0] == SPACEx4
	  && *(uint32 *) &p->name[ 4] == SPACEx4
	  && *(uint32 *) &p->name[ 8] == SPACEx4
	  && *(uint32 *) &p->name[12] == SPACEx4);
}

#ifndef GEOMETRY_TYPE
# define GEOMETRY_TYPE	struct disk_geometry
#endif

extern int open_partition (void); /* in disk_io.c */

/* Show details of the device to user.  */
static void
print_device (void)
{
  unsigned int const cylinder_size
    = NSECTORS_SIZE (&DATA.saved_geometry, DATA.sectors_per_cylinder);

  printf (MSG2 ("\n\
Device (%s): DA/UA=0x%2x (%u cylinders/%u heads/%u sectors)\n\
  %u bytes/sector  %u.%uKB/cylinder  total %u KB\n\n", "\n\
ǥХ (%s): DA/UA=0x%02x (%u /%u إå/%u )\n\
  %uХ/  %u.%uKB/   %uKB\n\n"),
	  DATA.device_name, DATA.saved_geometry.da_ua, DATA.cylinders,
	  DATA.saved_geometry.heads, DATA.saved_geometry.sectors,
	  SECTOR_SIZE (&DATA.saved_geometry),
	  cylinder_size >> 10, ((cylinder_size & 1023) * 10) >> 10,
	  (unsigned int) (((uint64) TOTAL_SECTORS (&DATA.saved_geometry)
			   << SECTOR_SIZE_SHIFT (&DATA.saved_geometry))
			  >> 10));
}

static int
pt_print (char *arg)
{
  int i;
  int any;
  const GEOMETRY_TYPE *const g = &DATA.saved_geometry;

  if (!begun ())
    return 1;

#define SECTORS_TO_KB(x) \
  ((uint32) (((uint64) (x) << SECTOR_SIZE_SHIFT (g)) >> 10))

  print_device ();

  for (i = 0, any = 0; i < DATA.n_partitions; i++)
    {
      const struct partition_descriptor *p = &DATA.partition_table[i];

      if (empty_partition_p (p))
	continue;

      if (!any)
	printf ("\
 No  	 (KB)	  ̾		    \n\
 --  --------  ------------  ----------------  ----------------\n");
      printf ("%3u%9u%14u   ", INDEX_BASE + i, p->start.c,
	      (p->end.c >= p->start.c
	       ? SECTORS_TO_KB ((p->end.c - p->start.c + 1)
				* (g->heads * g->sectors))
	       : 0));
      print_partition_name (p->name);
      putchar (' ');
      putchar (' ');
      printf (partition_type_string (p->id) ?: "%2x %2x",
	      p->id & 255, p->id >> 8);
      putchar ('\n');
      any = 1;
    }
  if (!any)
    printf (MSG2 ("Partition table is empty\n",
		  "ѡƥơ֥϶Ǥ\n"));
  return 0;
}

static int
pt_dump (char *arg)
{
  int i;

  printf (MSG2 ("\
 No  Type     IPL C/H/S     Start C/H/S     End C/H/S	       Name\n\
 --  -----  -------------  -------------  -------------  ----------------\n",
		"\
 No        IPL C/H/S      C/H/S	    λC/H/S	       ̾\n\
 --  -----  -------------  -------------  -------------  ----------------\n"));
  for (i = 0; i < DATA.n_partitions; i++)
    {
      const struct partition_descriptor *p = &DATA.partition_table[i];

      printf ("%3u  %2x %2x", INDEX_BASE + i, p->id & 255, p->id >> 8);
      printf ("  %5u/%3u/%3u", p->ipl.c, p->ipl.h, p->ipl.s);
      printf ("  %5u/%3u/%3u", p->start.c, p->start.h, p->start.s);
      printf ("  %5u/%3u/%3u", p->end.c, p->end.h, p->end.s);
      putchar (' ');
      putchar (' ');
      print_partition_name (p->name);
      putchar ('\n');
    }
  return 0;
}

static void
pt_warn (n, fmt, arg1, arg2, arg3, arg4)
     int n;
     const char *fmt;
     void *arg1, *arg2, *arg3, *arg4;
{
  printf (MSG2 ("Partition %2u: ","ê %2u: "), INDEX_BASE + n);

  printf (fmt, arg1, arg2, arg3, arg4);
  putchar ('\n');
}

#define BDA_OK(bda)	((bda)->c < DATA.cylinders	      && \
			 (bda)->h < DATA.saved_geometry.heads && \
			 (bda)->s < DATA.saved_geometry.sectors)

static int
pt_check (char *arg)
{
  int i;
  int not_sorted;
  uint16 largest_cylinder;

  if (!begun ())
    return 1;

  largest_cylinder = 0;
  for (i = 0, not_sorted = 0; i < DATA.n_partitions; i++)
    {
      int j;
      const struct partition_descriptor *p = &DATA.partition_table[i];

      if (p->_unused)
	pt_warn (i, MSG2 ("unused area of this entry contains a junk value",
			  "ȥ̤ΰ˥ߤäƤޤ"));

      if (p->id == 0)
	goto check_name;	/* Skip checking disk-addresses.  */

      if (!BDA_OK (&p->start))
	pt_warn (i, MSG2 ("%s address (%u/%u/%u) is out of range",
			  "%sɥ쥹 (%u/%u/%u) ϰϳǤ"),
		 MSG2 ("start", ""), p->start.c, p->start.h, p->start.s);
      else if (p->start.h || p->start.s)
	pt_warn (i, MSG2 ("\
%s head/sector address (%u/%u) will be ignroed", "\
%sɥ쥹Υإå/ (%u/%u) ̵뤵ޤ"),
		 MSG2 ("start", ""), p->start.h, p->start.s);

      if (p->end.c < p->start.c)
	pt_warn (i, MSG2 ("starting cylinder (%u) is larger than ending (%u)",
			  "ϥ (%u) λ (%u) 礭ʤäƤޤ"),
		 p->start.c, p->end.c);
      else if (!BDA_OK (&p->end))
	pt_warn (i, MSG2 ("%s address (%u/%u/%u) is out of range",
			  "%sɥ쥹 (%u/%u/%u) ϰϳǤ"),
		 MSG2 ("end", "λ"), p->end.c, p->end.h, p->end.s);
      else if ((p->end.h != 0 && p->end.h != DATA.saved_geometry.heads - 1) ||
	       (p->end.s != 0 && p->end.s != DATA.saved_geometry.sectors - 1))
	pt_warn (i, MSG2 ("\
%s head/sector address (%u/%u) will be ignroed", "\
%sɥ쥹Υإå/ (%u/%u) ̵뤵ޤ"),
		 MSG2 ("end", "λ"), p->start.h, p->start.s);

      if (!BDA_OK (&p->ipl))
	pt_warn (i, MSG2 ("%s address (%u/%u/%u) is out of range",
			  "%sɥ쥹 (%u/%u/%u) ϰϳǤ"),
		 "IPL", p->ipl.c, p->ipl.h, p->ipl.s);
      else if (p->start.c <= p->end.c
	       && (p->ipl.c < p->start.c || p->end.c < p->ipl.c))
	pt_warn (i, MSG2 ("IPL address (%u/%u/%u) is out of the partition",
			  "IPLɥ쥹 (%u/%u/%u) ѡƥ󳰤Ǥ"),
		 p->ipl.c, p->ipl.h, p->ipl.s);

      for (j = 0; j < i; j++)
	{
	  const struct partition_descriptor *q = &DATA.partition_table[j];

	  if (p->id != 0 && p->start.c <= q->end.c && q->start.c <= p->end.c)
	    pt_warn (i, MSG2 ("overlapping with partition %u",
			      "ѡƥ %u ȽŤʤäƤޤ"),
		     INDEX_BASE + j);
	}

      if (largest_cylinder < p->start.c)
	largest_cylinder = p->start.c;
      else if (largest_cylinder > p->start.c)
	not_sorted = 1;

    check_name:
      for (j = 0; j < PARTITION_NAME_SIZE; j++)
	if (p->name[j] < ' ' || p->name[j] == '\x7F'
	    || p->name[j] == (uint8) '\x80' || p->name[j] == (uint8) '\xA0'
	    || (uint8) '\xFC' < p->name[j])
	  break;
	else if ((((uint8) '\x80' <= p->name[j]
		   && p->name[j] <= (uint8) '\x9F')
		  || (uint8) '\xE0' <= p->name[j])
		 && (j >= PARTITION_NAME_SIZE - 1
		     || p->name[++j] < '\x40' || p->name[j] != '\x7F'
		     || (uint8) '\xFC' < p->name[j]))
	  break;
      if (j < PARTITION_NAME_SIZE)
	pt_warn (i, MSG2 ("\
partition name's byte #%u `\\x%2x' is not printable", "\
ѡƥ̾ %u Х `\\x%2x' ɽǤޤ"), j, p->name[j]);
    }
  if (not_sorted)
    printf (MSG2 ("Partitions not properly sorted\n",
		  "ѡƥ󤬥ˤʤäƤޤ\n"));
  return 0;
}

static struct partition_descriptor *
parse_partition_number (char **argp)
{
  int n;

  if (!begun ())
    return NULL;
  if (!safe_parse_maxint (argp, &n))
    return NULL;
  if (**argp && !isblank (**argp))
    {
      errnum = ERR_NUMBER_PARSING;
      return NULL;
    }
  do
    (*argp)++;
  while (isblank (**argp));
  if (n < INDEX_BASE || INDEX_BASE + DATA.n_partitions <= n)
    {
      errnum = ERR_NO_PART;
      return NULL;
    }
  return DATA.partition_table + (n - INDEX_BASE);
}

static int
pt_read (char *arg)
{
  unsigned int size = SECTOR_SIZE (&DATA.saved_geometry);
  char buf[size];

  if (!begun () && !overwrite_ok ())
    return 1;

  if (!open (arg))
    return 1;
  if (read (buf, size) < size)
    {
      if (!errnum)
	errnum = ERR_BAD_ARGUMENT; /* XXX */
      return 1;
    }

  memmove (DATA.partition_table, buf, size);
  DATA.edited = 1;
  return 0;
}
  
#define SET_CHS(dba, C, H, S)					\
	do { register uint32 _chs = (C) << 16;			\
	     if (__builtin_constant_p ((H)))			\
	       _chs |= ((H) & 255U) << 8;			\
	     else						\
	       __asm ("mov%B0\t%b1,%h0"				\
		      : "=q" (_chs)				\
		      : "iqm" ((uint8) (H)), "0" (_chs));	\
	     if (__builtin_constant_p ((S)))			\
	       _chs |= (S) & 255U;				\
	     else						\
	       __asm ("mov%B0\t%b1,%b0"				\
		      : "=q" (_chs)				\
		      : "iqm" ((uint8) (S)), "0" (_chs));	\
	     *(uint32 *) (dba) = _chs; } while (0)

static int
pt_region (char *arg)
{
  int st, en;
  char c;
  struct partition_descriptor *p;

  if (!(p = parse_partition_number (&arg)))
    return 1;
  if (*arg == '-' || *arg == '+')
    {
      int i;

      st = 1;
    next_st:
      for (i = 0; i < DATA.n_partitions; i++)
	if (DATA.partition_table[i].id
	    && DATA.partition_table[i].start.c <= st
	    && st <= DATA.partition_table[i].end.c)
	  {
	    st = DATA.partition_table[i].end.c + 1;
	    goto next_st;
	  }
      if (st >= DATA.cylinders)
	{
	  printf (MSG2 ("No more unused region\n",
			"ΰ褬Ĥޤ\n"));
	  BAD_ARGUMENT;
	}
    }
  else if (!safe_parse_maxint (&arg, &st))
    return 1;
  else if (st > 0xFFFF)
    BAD_ARGUMENT;
  if ((c = *arg++) != '-' && c != '+')
    BAD_ARGUMENT;

  if (!*arg || isblank (*arg))
    {
      int i;
      for (i = 0, en = DATA.cylinders - 1; i < DATA.n_partitions; i++)
	if (DATA.partition_table[i].id
	    && st < DATA.partition_table[i].start.c
	    && DATA.partition_table[i].start.c <= en)
	  en = DATA.partition_table[i].start.c - 1;
    }
  else if (!safe_parse_maxint (&arg, &en))
    return 1;
  else if (c == '+'
	   && (*arg == 'k' || *arg == 'K' || *arg == 'm' || *arg == 'M'
	       || *arg == 'g' || *arg == 'G' || *arg == 's'))
    {
      int shift = (*arg == 's'
		   ? 0 : (((*arg == 'g' || *arg == 'G') ? 30 :
			   (*arg == 'm' || *arg == 'M') ? 20 : 10)
			  - SECTOR_SIZE_SHIFT (&DATA.saved_geometry)));

      if (shift > 0)
	{
	  if ((unsigned int) en > (~0U >> shift))
	    ERROR_RETURN (ERR_NUMBER_PARSING);

	  en <<= shift;
	}
      else
	en >>= -shift;

      en = CEIL_UDIV (en, DATA.sectors_per_cylinder);
      arg++;
    }
  if (*arg && !isblank (*arg))
    ERROR_RETURN (ERR_NUMBER_PARSING);

  if (en > 0xFFFF || (c == '+' && (en = st + en - 1) > 0xFFFF))
    {
      printf (MSG2 ("Ending cylinder exceeds 16-bit limit\n",
		    "λ16ӥåȤ˼ޤޤ\n"));
      BAD_ARGUMENT;
    }
  if (st > en)
    {
      printf (MSG2 ("\
Ending cyliner must not be smaller than starting cylinder\n", "\
λ򳫻ϥ꾮뤳ȤϤǤޤ\n"));
      BAD_ARGUMENT;
    }
  SET_CHS (&p->ipl,	st, 0, 0);
  SET_CHS (&p->start,	st, 0, 0);
  SET_CHS (&p->end,
	   en, DATA.saved_geometry.heads - 1, DATA.saved_geometry.sectors - 1);
  DATA.edited = 1;
  return 0;
}

static int
pt_clear (char *arg)
{
  struct partition_descriptor *p;

  if (!(p = parse_partition_number (&arg)))
    return 1;
  ((uint32 *) p)[0] = 0;
  ((uint32 *) p)[1] = 0;
  ((uint32 *) p)[2] = 0;
  ((uint32 *) p)[3] = 0;
  ((uint32 *) p)[4] = SPACEx4;
  ((uint32 *) p)[5] = SPACEx4;
  ((uint32 *) p)[6] = SPACEx4;
  ((uint32 *) p)[7] = SPACEx4;
  DATA.edited = 1;
  return 0;
}

static int
pt_name (char *arg)
{
  int i;
  struct partition_descriptor *p;

  if (!(p = parse_partition_number (&arg)))
    return 1;
  if (!*arg)
    {
      printf (MSG2 ("Blank partition name not allowed\n",
		    "ѡƥ̾ˤ뤳ȤϤǤޤ\n"));
      BAD_ARGUMENT;
    }
  for (i = 0; i < PARTITION_NAME_SIZE; i++)
    p->name[i] = *arg ? *arg++ : ' ';
  DATA.edited = 1;
  return 0;
}

static int
pt_type (char *arg)
{
  struct partition_descriptor *p;
  int mid, sid;

  if (!*arg)
    {
      const uint8 *kp;

      printf (MSG2 ("Known partition types:\n", "ѡƥ󥿥:\n"));
      for (kp = known_type; *kp; kp++)
	{
	  printf ("  0x%2x 0x%2x  %s\n", kp[0], kp[1], &kp[2]);
	  for (++kp; *++kp; /* nop */);
	}
      return 0;
    }

  if (!(p = parse_partition_number (&arg)))
    return 1;
  if (!safe_parse_maxint (&arg, &mid))
    return 1;
  while (*arg && isblank (*arg))
    arg++;
  if (!safe_parse_maxint (&arg, &sid))
    return 1;
  if (*arg && !isblank (*arg))
    ERROR_RETURN (ERR_NUMBER_PARSING);

  if (mid > 0xFF || sid >= 0xFF)
    BAD_ARGUMENT;
  ((uint8 *) &p->id)[0] = mid;
  ((uint8 *) &p->id)[1] = sid;
  DATA.edited = 1;
  return 0;
}

static int
pt_swap (char *arg)
{
  struct partition_descriptor *p, *q;
  struct partition_descriptor tmp;

  if (!(p = parse_partition_number (&arg)) ||
      !(q = parse_partition_number (&arg)))
    return 1;
  if (p == q)
    return 0;
  memmove (&tmp, p, sizeof (struct partition_descriptor));
  memmove (p, q, sizeof (struct partition_descriptor));
  memmove (q, &tmp, sizeof (struct partition_descriptor));
  DATA.edited = 1;
  return 0;
}

/* Partition discovery */

#define PAGEBUF		((char *) FSYS_BUF)
#define PAGESIZE	4096

struct linux_swap_info		/* from <linux/swap.h> */
  {
    char bootbits[1024];
    unsigned int version;
    unsigned int last_page;
    unsigned int nr_badpages;
    unsigned int padding[125];
    unsigned int badpages[1];
  };

#define V1SWAP		((struct linux_swap_info *) PAGEBUF)

#define __u8	uint8
#define __s16	int16
#define __u16	uint16
#define __s32	int32
#define __u32	uint32

struct ext2_super_block
  {
    __u32 s_inodes_count;	/* Inodes count */
    __u32 s_blocks_count;	/* Blocks count */
    __u32 s_r_blocks_count;	/* Reserved blocks count */
    __u32 s_free_blocks_count;	/* Free blocks count */
    __u32 s_free_inodes_count;	/* Free inodes count */
    __u32 s_first_data_block;	/* First Data Block */
    __u32 s_log_block_size;	/* Block size */
    __s32 s_log_frag_size;	/* Fragment size */
    __u32 s_blocks_per_group;	/* # Blocks per group */
    __u32 s_frags_per_group;	/* # Fragments per group */
    __u32 s_inodes_per_group;	/* # Inodes per group */
    __u32 s_mtime;		/* Mount time */
    __u32 s_wtime;		/* Write time */
    __u16 s_mnt_count;		/* Mount count */
    __s16 s_max_mnt_count;	/* Maximal mount count */
    __u16 s_magic;		/* Magic signature */
    __u16 s_state;		/* File system state */
    __u16 s_errors;		/* Behaviour when detecting errors */
    __u16 s_pad;
    __u32 s_lastcheck;		/* time of last check */
    __u32 s_checkinterval;	/* max. time between checks */
    __u32 s_creator_os;		/* OS */
    __u32 s_rev_level;		/* Revision level */
    __u16 s_def_resuid;		/* Default uid for reserved blocks */
    __u16 s_def_resgid;		/* Default gid for reserved blocks */

    __u32 s_first_ino;
    __u16 s_inode_size;
    __u16 s_block_group_nr;

    __u32 s_feature_compat;
    __u32 s_feature_incompat;
    __u32 s_feature_ro_compat;

    __u8  s_uuid[16];
    char  s_volume_name[16];
    char  s_last_mounted[64];
    __u32 s_algorithm_usage_bitmap;
    __u32 s_prealloc_dir_blocks;
  };

#define EXT2_SUPER_MAGIC	0xEF53
#define EXT2_MIN_BLOCK_SIZE	1024
#define EXT2_BLOCK_SIZE(s)	(EXT2_MIN_BLOCK_SIZE << (s)->s_log_block_size)
#define EXT2SB			((struct ext2_super_block *) (PAGEBUF + 1024))

#define _PACKED		__attribute__ ((packed))

struct fat_boot_sector
  {
    __u8  jump_code[3];		/* bootstrap short or near jump */
    __u8  system_id[8];		/* OEM name */
    /* BIOS Parameter Block (BPB) */
    __u16 sector_size;		/* bytes/sector (usually 512 or 2048) */
    __u8  cluster_size;		/* # of sectors/cluster */
    __u16 reserved;		/* # of reserved sectors */
    __u8  fats;			/* # of copies of FAT (usually 2) */
    __u16 dir_entries;		/* max # of root directory entries */
                                /* (usually 512; 0 for FAT32) */
    __u16 sectors;		/* total # of sectors (if part < 32MB) or 0 */
    __u8  media;		/* media code (0xf8 for hard disks) */
    __u16 fat_length;		/* # of sectors/FAT */
    __u16 secs_track;		/* # of sectors/track (in 1..63) */
    __u16 heads;		/* # of heads (in 1..255) */
    __u32 hidden; 		/* # of hidden sectors in partition */
    __u32 sectors32;		/* total # of sectors (if sectors == 0) */
  } _PACKED;

#define FATBS			((struct fat_boot_sector *) PAGEBUF)
/* `jump_code' jumps over at least this displacement.  */
#define FATBS_MIN_JUMP \
	((int) ((char *) &((struct fat_boot_sector *) 0)->sectors32 \
		- (char *) 0))

/* Extended BIOS parameter block */
#define EXTENDED_BPB_MEMBERS						    \
    __u8  drive_number;		/* logical drive number of partition */	    \
    __u8  current_head;		/* track containing boot record */	    \
    __u8  extended_signature;	/* extended signature (= 0x29) */	    \
    __u8  serial_nr[4];		/* serial number of partition */	    \
    __u8  volume_name[11];	/* volume name of partition */		    \
    __u8 fs_name[8]   		/* filesystem name (FAT12, FAT16, FAT32) */

struct fat16_boot_sector
  {
    struct fat_boot_sector s;
    EXTENDED_BPB_MEMBERS;
  } _PACKED;

#define FAT16BS			((struct fat16_boot_sector *) PAGEBUF)

struct fat32_boot_sector
  {
    struct fat_boot_sector s;

    __u32 fat32_length;		/* sectors/FAT */
    __u16 flags;		/* bit 7: fat mirroring, low 4: active fat */
    __u8  version[2];		/* major, minor filesystem version */
    __u32 root_cluster;		/* first cluster in root directory */
    __u16 info_sector;		/* filesystem info sector # relative
                                   to partition start (usually 1) */
    __u16 backup_boot;		/* backup boot sector */
    __u16 reserved2[6];		/* Unused */

    EXTENDED_BPB_MEMBERS;
  } _PACKED;

#define FAT32BS			((struct fat32_boot_sector *) PAGEBUF)

struct bsd_partition
  {
    __u32 p_size;		/* number of sectors in partition */
    __u32 p_offset;		/* starting sector */
    __u32 p_fsize;		/* filesystem basic fragment size */
    __u8  p_fstype;		/* filesystem type, see below */
    __u8  p_frag;		/* filesystem fragments per block */
    __u16 p_cpg;		/* filesystem cylinders per group */
  };

#define BSD_DISKMAGIC		(0x82564557UL)
#define BSD_RAW_PART		2

struct bsd_disklabel
  {
    __u32 d_magic;		/* the magic number */
    __s16 d_type;		/* drive type */
    __s16 d_subtype;		/* controller/d_type specific */
    char  d_typename[16];	/* type name, e.g. "eagle" */
    char  d_packname[16];	/* pack identifier */ 
    __u32 d_secsize;		/* # of bytes per sector */
    __u32 d_nsectors;		/* # of data sectors per track */
    __u32 d_ntracks;		/* # of tracks per cylinder */
    __u32 d_ncylinders;		/* # of data cylinders per unit */
    __u32 d_secpercyl;		/* # of data sectors per cylinder */
    __u32 d_secperunit;		/* # of data sectors per unit */
    __u16 d_sparespertrack;	/* # of spare sectors per track */
    __u16 d_sparespercyl;	/* # of spare sectors per cylinder */
    __u32 d_acylinders;		/* # of alt. cylinders per unit */
    __u16 d_rpm;		/* rotational speed */
    __u16 d_interleave;		/* hardware sector interleave */
    __u16 d_trackskew;		/* sector 0 skew, per track */
    __u16 d_cylskew;		/* sector 0 skew, per cylinder */
    __u32 d_headswitch;		/* head switch time, usec */
    __u32 d_trkseek;		/* track-to-track seek, usec */
    __u32 d_flags;		/* generic flags */
#define BSD_NDDATA	5
    __u32 d_drivedata[BSD_NDDATA]; /* drive-type specific information */
#define BSD_NSPARE 5
    __u32 d_spare[BSD_NSPARE];	/* reserved for future use */
    __u32 d_magic2;		/* the magic number (again) */
    __u16 d_checksum;		/* xor of data incl. partitions */

    /* filesystem and partition information: */
    __u16 d_npartitions;	/* number of partitions in following */
    __u32 d_bbsize;		/* size of boot area at sn0, bytes */
    __u32 d_sbsize;		/* max size of fs superblock, bytes */
    struct bsd_partition d_partitions[8]; /* actually may be more */
  };

#define BSDLABEL		((struct bsd_disklabel *) (PAGEBUF + 512))

#define PARTTABLE	((struct partition_descriptor *) (FSYS_BUF + PAGESIZE))

static int
pt_guess (char *arg)
{
  unsigned int i;
  unsigned int cyl;
  unsigned int nsect;
  unsigned int npart;
  unsigned int first_unused_cyl;
  unsigned int sector_size;

  if (!begun () || !overwrite_ok ())
    return 1;

  current_drive = DATA.saved_geometry.da_ua;
  current_partition = 0xFFFFFF;	/* Use whole disk. */

  if (!open_partition ())
    return 0;

  for (i = 0; i < DATA.n_partitions; i++)
    {
      uint32 *const p = (uint32 *) &PARTTABLE[i];
      p[0] = p[1] = p[2] = p[3] = MAKE_U32 (0, 0, 0, 0);
      p[4] = p[5] = p[6] = p[7] = MAKE_U32 (' ', ' ', ' ', ' ');
    }

  print_device ();
  sector_size = SECTOR_SIZE (&DATA.saved_geometry);

  npart = 0;
  for (cyl = first_unused_cyl = 1; cyl < DATA.cylinders; cyl++)
    {
      uint16 type;
      const char *desc;
      unsigned int const sect = cyl * DATA.sectors_per_cylinder;

      putchar_x = 0;
      printf (" %5u: ", cyl);
      if (ASCII_CHAR (checkkey ()) == 3)	/* ^C is pressed */
	{
	  printf ("()\n");
	  return 0;
	}
      if (!rawread (current_drive, sect, 0, PAGESIZE, PAGEBUF))
	{
	  print_error ();
	  continue;
	}

      if (*(uint32 *) (PAGEBUF + PAGESIZE - 10)
	  == MAKE_U32 ('S', 'W', 'A', 'P'))
	{
	  if (*(uint32 *) (PAGEBUF + PAGESIZE - 6)
	      == MAKE_U32 ('-', 'S', 'P', 'A'))
	    {
	      if (*(uint16 *) (PAGEBUF + PAGESIZE - 2) == MAKE_U16 ('C', 'E'))
		{
		  /* Linux version 0 swap */
		  int i = PAGESIZE - 10;
		  int j;
		  while (!PAGEBUF[i - 1] && --i)
		    ;
		  for (j = 7; j >= 0 && !(PAGEBUF[i - 1] && (1 << j)); --j)
		    ;
		  nsect = (i - 1) * 8 + j + 1;
		  type = PC_SLICE_TYPE_LINUX_SWAP;
		  desc = "Linux v0 åΰ";
		  goto add_partition;
		}
	    }
	  else if (*(uint32 *) (PAGEBUF + PAGESIZE - 6)
		   == MAKE_U32 ('S', 'P', 'A', 'C'))
	    {
	      if (*(uint16 *) (PAGEBUF + PAGESIZE - 2) == MAKE_U16 ('E', '2')
		  && V1SWAP->version == 1)
		{
		  /* Linux version 1 swap */
		  nsect = (V1SWAP->last_page + 1)
		    * TO_NR_SECTORS (&DATA.saved_geometry, PAGESIZE);
		  type = PC_SLICE_TYPE_LINUX_SWAP;
		  desc = "Linux v1 åΰ";
		  goto add_partition;
		}
	    }
	}

      if (EXT2SB->s_magic == EXT2_SUPER_MAGIC
#if 0
	  && EXT2SB->s_blocks_count > EXT2SB->s_r_blocks_count
	  && EXT2SB->s_blocks_count > EXT2SB->s_free_blocks_count
	  && EXT2SB->s_inodes_count > EXT2SB->s_free_inodes_count
#endif
	  && EXT2SB->s_log_block_size <= 2 /* 1k, 2k, 4k block for 0-2 */
	  && EXT2SB->s_rev_level <= 1 /* 0:GOOD_OLD_REV, 1:DYNAMIC_REV */
	  && EXT2SB->s_block_group_nr == 0)
	{
	  nsect = EXT2SB->s_blocks_count
	    * TO_NR_SECTORS (&DATA.saved_geometry, EXT2_BLOCK_SIZE (EXT2SB));
#if 0
	  printf ("ext2: %u ֥å (%u ), ֥å %u\n",
		  EXT2SB->s_blocks_count, nsect, EXT2_BLOCK_SIZE (EXT2SB));
#endif
	  type = PC_SLICE_TYPE_EXT2FS;
	  desc = "ext2 ե륷ƥ";
	  goto add_partition;
	}

      if (BSDLABEL->d_magic == BSD_DISKMAGIC
	  && BSDLABEL->d_magic2 == BSD_DISKMAGIC
	  && BSDLABEL->d_npartitions > BSD_RAW_PART)
	{
	  nsect = BSDLABEL->d_partitions[BSD_RAW_PART].p_size;
	  type = PC_SLICE_TYPE_BSD;
	  desc = "BSDǥ٥";
	  goto add_partition;
	}

      if (*((uint16 *) (PAGEBUF + sector_size) - 1) == 0xAA55
	  && FATBS->sector_size >= sector_size
	  && FATBS->sector_size <= 4096
	  /* Check FATBS->sector_size is exact power of 2.  */
	  && (FATBS->sector_size & (FATBS->sector_size - 1)) == 0
	  && ((FATBS->jump_code[0] == 0xE9 /* JMP NEAR */
	       && ({ int16 const target = 3 + *(int16 *) &FATBS->jump_code[1];
		     target == 0 || (FATBS_MIN_JUMP <= target
				     && target < FATBS->sector_size); }))
	      || (FATBS->jump_code[0] == 0xEB /* JMP SHORT */
		  && ((FATBS_MIN_JUMP - 2 <= FATBS->jump_code[1]
		       && FATBS->jump_code[1] < 0x80)
		      || FATBS->jump_code[1] == 0xFE)))
	  && FATBS->cluster_size != 0
	  && FATBS->media == 0xF8)
	{
	  /* DOS boot sector? */

	  if (FAT16BS->extended_signature == 0x29
	      && (*(uint32 *) &FAT16BS->fs_name
		  == MAKE_U32 ('F', 'A', 'T', '1')))
	    {
	      nsect = FATBS->sectors
		* TO_NR_SECTORS (&DATA.saved_geometry, FATBS->sector_size);

	      switch (FAT16BS->fs_name[4])
		{
		case '2':
		  type = PC_SLICE_TYPE_FAT12;
		  desc = "FAT12 ե륷ƥ";
		  goto add_partition;
		case '6':
		  if (FATBS->sectors)
		    {
		      type = PC_SLICE_TYPE_FAT16_LT32M;
		      desc = "FAT16 ե륷ƥ (32MB̤)";
		    }
		  else
		    {
		      nsect = FATBS->sectors32
			* TO_NR_SECTORS (&DATA.saved_geometry,
					 FATBS->sector_size);
		      type = PC_SLICE_TYPE_FAT16_GT32M;
		      desc = "FAT16 ե륷ƥ (32MBʾ)";
		    }
		  goto add_partition;
		}
	    }

	  if (FAT32BS->extended_signature == 0x29
	      && (((uint32 *) &FAT32BS->fs_name)[0]
		  == MAKE_U32 ('F', 'A', 'T', '3'))
	      && (((uint32 *) &FAT32BS->fs_name)[1]
		  == MAKE_U32 ('2', ' ', ' ', ' ')))
	    {
	      nsect = FATBS->sectors32
		* TO_NR_SECTORS (&DATA.saved_geometry, FATBS->sector_size);
	      type = PC_SLICE_TYPE_FAT32;
	      desc = "FAT32 ե륷ƥ";
	      goto add_partition;
	    }

	  if (*(uint32 *) &FATBS->system_id[0] == MAKE_U32 ('N', 'T', 'F', 'S')
	      && (*(uint32 *) &FATBS->system_id[4]
		  == MAKE_U32 (' ', ' ', ' ', ' '))
	      && !FATBS->fats && !FATBS->dir_entries)
	    {
	      /* XXX: PAGEBUF + 0x2C holds upper 32-bits of # clusters */
	      nsect = *(uint32 *) (PAGEBUF + 0x28)
		* TO_NR_SECTORS (&DATA.saved_geometry, FATBS->sector_size);
	      /* Count backup boot sector if any. */
	      if (rawread (current_drive, sect + nsect, 0, 512, PAGEBUF + 512)
		  && !__builtin_memcmp (PAGEBUF, PAGEBUF + 512, 512))
		++nsect;
	      type = PC_SLICE_TYPE_FAT32 /* PC_SLICE_TYPE_NTFS */;
	      desc = "NTFS ܥ塼";
	      goto add_partition;
	    }
	}

      /* Maybe check UFS superblock (offset is 8k), Reiserfs (8k or 64k)
	 here... */

      continue;
    add_partition:
      {
	unsigned int ncyl = CEIL_UDIV (nsect, DATA.sectors_per_cylinder);

	if (sect + nsect >= TOTAL_SECTORS (&DATA.saved_geometry)
	    || sect + nsect < sect || !ncyl)
	  continue;
	if (cyl > first_unused_cyl)
	  {
	    putchar_x = 0;
	    printf (" %5u: ", first_unused_cyl);
	    printf ("( %d )\n", cyl - first_unused_cyl);
	    printf (" %5u: ", cyl);
	  }
	printf ("%s (%d ) ", desc, ncyl);
	if (npart < DATA.n_partitions)
	  {
	    /* PARTTABLE is already cleared above, so we need to write
	       only non-zero (or non-blank for `name') fields.  */
	    PARTTABLE[npart].id = type;
	    PARTTABLE[npart].ipl.c = cyl;
	    PARTTABLE[npart].start.c = cyl;
	    SET_CHS (&PARTTABLE[npart].end, cyl + ncyl - 1,
		     DATA.saved_geometry.heads - 1,
		     DATA.saved_geometry.sectors - 1);
	    *(uint32 *) &PARTTABLE[npart].name[0]
	      = MAKE_U32 ('P', 'a', 'r', 't');
	    *(uint32 *) &PARTTABLE[npart].name[4]
	      = MAKE_U32 ('i', 't', 'i', 'o');
	    PARTTABLE[npart].name[8] = 'n';
	    convert_to_ascii (PARTTABLE[npart].name + 10, 'u',
			      INDEX_BASE + npart);

	    printf (MSG2 ("partition %d", "ѡƥ %d"), npart);
	  }
	else
	  printf (MSG2 ("*", ""));
	npart++;
	printf ("\n");
	cyl += ncyl - 1;	/* -1 to cancel for's loop expression */
	first_unused_cyl = cyl + 1;
      }
    }

  putchar_x = 0;
  if (first_unused_cyl < DATA.cylinders)
    {
      printf (" %5u: ", first_unused_cyl);
      printf ("( %d )\n", DATA.cylinders - first_unused_cyl);
    }
  else
    printf ("       ");
  putchar ('\n');

  if (npart > DATA.n_partitions)
    printf (MSG2 ("\
WARNING: more than maximum (%d) partitions found (marked *)\n", "\
ٹ:  (%d) Ķѡƥ󤬸Ĥޤ ()\n"),
	    DATA.n_partitions);
  memmove (DATA.partition_table, PARTTABLE,
	   sizeof (struct partition_descriptor) * DATA.n_partitions);
  DATA.edited = 1;
  return 0;
}

struct subcommand
  {
    char *name;
    int (*func) (char *);
  };

static const struct subcommand pt_subcommand[] = {
  { "begin",	pt_begin },
  { "end",	pt_end },
  { "write",	pt_write },
  { "print",	pt_print },
  { "dump",	pt_dump },
  { "read",	pt_read },
  { "region",	pt_region },
  { "clear",	pt_clear },
  { "name",	pt_name },
  { "type",	pt_type },
  { "swap",	pt_swap },
  { "check",	pt_check },
  { "guess",	pt_guess },
};

#define ARRAY_SIZE(array)	(sizeof (array) / sizeof (array[0]))

#define PT_HELP_STRING	MSG2 ("\
pt: Trivial partition table editor\n\
\n\
Usage: pt SUBCOMMAND [ARG...]\n\
\n\
Subcommands:\n\
  begin DEVICE		read partition table from DEVICE and begin editing\n\
  write			write the partition table back to the disk\n\
  end			throw away on-memory partition table\n\
  print			list partitions\n\
  dump			dump partition table in raw form\n\
\n\
  read FILE		read partition table from FILE\n\
  clear N		clear partition N\n\
  region N REGION	set region for partition N\n\
  name N NAME		set name for partition N\n\
  type [N MID SID]	set type for partition N\n\
  check			perform simple consistency check of partition table\n\
  guess			try to rebuild partition table by probing disk\n\
\n\
REGION may be in the form of START-END or START+SIZE.  START and END are\n\
in unit of cylinders.  SIZE is also in cylinders by default, but can be\n\
in sectors, KB, MB or GB with suffix `s', `k', `m' and `g' respectively.\n\
START, END and SIZE can be omitted.\n", "\
pt: ʰץѡƥơ֥륨ǥ\n\
\n\
ˡ: pt ֥ޥ [...]\n\
\n\
֥ޥ:\n\
  begin <ǥХ>	<ǥХ> Υѡƥơ֥Խ򳫻Ϥ\n\
  write			ѡƥơ֥ǥ˽᤹\n\
  end			ԽΥѡƥơ֥˴\n\
  print			ԽΥѡƥơ֥ɽ\n\
  dump			ԽΥѡƥơ֥ܺ٤ɽ\n\
\n\
  read <ե>	<ե> ѡƥơ֥ɤ߹\n\
  clear N		ѡƥ N ä\n\
  region N <ΰ>	ѡƥ N ΰꤹ\n\
  name N <̾>		ѡƥ N ̾ꤹ\n\
  type [N <>]	ѡƥ N Υפꤹ\n\
  check			ѡƥơ֥δñʥåԤ\n\
  guess			ǥƤѡƥơ֥ߤ\n\
\n\
<ΰ>  <ϥ>-<λ>	ޤ <ϥ>+<> \n\
ꤷƤ <> ϥñ̤Ǥ skmg Τ줫\n\
ĤȤ줾쥻KBMBGB ñ̤ǻǤޤ\n\
<ϥ><λ><> ϤάǽǤ\n")

int
pt_func (char *arg)
{
  const struct subcommand *cmd, *ambig;

  if (!arg[0])
    {
      printf (PT_HELP_STRING);
      return 0;
    }

  for (cmd = pt_subcommand, ambig = NULL;
       cmd < pt_subcommand + ARRAY_SIZE (pt_subcommand);
       cmd++)
    {
      int i;
      for (i = 0; cmd->name[i] && arg[i] == cmd->name[i]; i++)
	;
      if (!arg[i] || isblank (arg[i]))
	{
	  if (!cmd->name[i])
	    goto go;
	  if (ambig)
	    {
	      ambig = NULL;
	      break;
	    }
	  ambig = cmd;
	}
    }
  if (!ambig)
    ERROR_RETURN (ERR_UNRECOGNIZED);

  cmd = ambig;
 go:
  while (*arg && !isblank (*arg))
    ++arg;
  while (*arg && isblank (*arg))
    ++arg;
  return cmd->func (arg);
}
